@@ -21,9 +21,9 @@ static const char *role_names [] =
   "dial",
   "dialog",
   "directory pane",
+  "drawing area",
   "file chooser",
   "filler",
-  "focus traversable",
   "font chooser",
   "frame",
   "glass pane",
@@ -62,12 +62,13 @@ static const char *role_names [] =
   "table column header",
   "table row header",
   "tearoff menu item",
+  "terminal",
   "text",
   "toggle button",
   "toolbar",
   "tooltip",
   "tree",
-  "tree-table",
+  "tree table",
   "<unknown>",
   "viewport",
   "window",
@@ -102,6 +103,7 @@ cspi_init_role_table (AccessibleRole *role_table)
   role_table [Accessibility_ROLE_DIAL] = SPI_ROLE_DIAL;
   role_table [Accessibility_ROLE_DIALOG] = SPI_ROLE_DIALOG;
   role_table [Accessibility_ROLE_DIRECTORY_PANE] = SPI_ROLE_DIRECTORY_PANE;
+  role_table[Accessibility_ROLE_DRAWING_AREA] = SPI_ROLE_DRAWING_AREA;
   role_table [Accessibility_ROLE_FILE_CHOOSER] = SPI_ROLE_FILE_CHOOSER;
   role_table [Accessibility_ROLE_FILLER] = SPI_ROLE_FILLER;
   role_table [Accessibility_ROLE_FONT_CHOOSER] = SPI_ROLE_FONT_CHOOSER;
@@ -139,14 +141,16 @@ cspi_init_role_table (AccessibleRole *role_table)
   role_table [Accessibility_ROLE_STATUS_BAR] = SPI_ROLE_STATUS_BAR;
   role_table [Accessibility_ROLE_TABLE] = SPI_ROLE_TABLE;
   role_table [Accessibility_ROLE_TABLE_CELL] = SPI_ROLE_TABLE_CELL;
+  role_table[Accessibility_ROLE_TABLE_COLUMN_HEADER] = SPI_ROLE_TABLE_COLUMN_HEADER;
+  role_table[Accessibility_ROLE_TABLE_ROW_HEADER] = SPI_ROLE_TABLE_ROW_HEADER;
   role_table [Accessibility_ROLE_TEAROFF_MENU_ITEM] = SPI_ROLE_TEAROFF_MENU_ITEM;
+  role_table[Accessibility_ROLE_TERMINAL] = SPI_ROLE_TERMINAL;
   role_table [Accessibility_ROLE_TEXT] = SPI_ROLE_TEXT;
   role_table [Accessibility_ROLE_TOGGLE_BUTTON] = SPI_ROLE_TOGGLE_BUTTON;
   role_table [Accessibility_ROLE_TOOL_BAR] = SPI_ROLE_TOOL_BAR;
   role_table [Accessibility_ROLE_TOOL_TIP] = SPI_ROLE_TOOL_TIP;
   role_table [Accessibility_ROLE_TREE] = SPI_ROLE_TREE;
   role_table [Accessibility_ROLE_TREE_TABLE] = SPI_ROLE_TREE_TABLE;
-  role_table [Accessibility_ROLE_UNKNOWN] = SPI_ROLE_UNKNOWN;
   role_table [Accessibility_ROLE_VIEWPORT] = SPI_ROLE_VIEWPORT;
   role_table [Accessibility_ROLE_WINDOW] = SPI_ROLE_WINDOW;
   role_table [Accessibility_ROLE_EXTENDED] = SPI_ROLE_EXTENDED;
@@ -177,6 +181,67 @@ cspi_role_from_spi_role (Accessibility_Role role)
   return cspi_role; 
 }
 
+static SPIBoolean
+init_state_table (Accessibility_StateType *state_table)
+{
+  int i;
+
+  for (i = 0; i < Accessibility_STATE_LAST_DEFINED; i++)
+    state_table[i] = SPI_STATE_INVALID;
+
+  state_table[SPI_STATE_ACTIVE] = Accessibility_STATE_ACTIVE;
+  state_table[SPI_STATE_ARMED] = Accessibility_STATE_ARMED;
+  state_table[SPI_STATE_BUSY] = Accessibility_STATE_BUSY;
+  state_table[SPI_STATE_CHECKED] = Accessibility_STATE_CHECKED;
+  state_table[SPI_STATE_DEFUNCT] = Accessibility_STATE_DEFUNCT;
+  state_table[SPI_STATE_EDITABLE] = Accessibility_STATE_EDITABLE;
+  state_table[SPI_STATE_ENABLED] = Accessibility_STATE_ENABLED;
+  state_table[SPI_STATE_EXPANDABLE] = Accessibility_STATE_EXPANDABLE;
+  state_table[SPI_STATE_EXPANDED] = Accessibility_STATE_EXPANDED;
+  state_table[SPI_STATE_FOCUSABLE] = Accessibility_STATE_FOCUSABLE;
+  state_table[SPI_STATE_FOCUSED] = Accessibility_STATE_FOCUSED;
+  state_table[SPI_STATE_HORIZONTAL] = Accessibility_STATE_HORIZONTAL;
+  state_table[SPI_STATE_ICONIFIED] = Accessibility_STATE_ICONIFIED;
+  state_table[SPI_STATE_MODAL] = Accessibility_STATE_MODAL;
+  state_table[SPI_STATE_MULTI_LINE] = Accessibility_STATE_MULTI_LINE;
+  state_table[SPI_STATE_MULTISELECTABLE] = Accessibility_STATE_MULTISELECTABLE;
+  state_table[SPI_STATE_OPAQUE] = Accessibility_STATE_OPAQUE;
+  state_table[SPI_STATE_PRESSED] = Accessibility_STATE_PRESSED;
+  state_table[SPI_STATE_RESIZABLE] = Accessibility_STATE_RESIZABLE;
+  state_table[SPI_STATE_SELECTABLE] = Accessibility_STATE_SELECTABLE;
+  state_table[SPI_STATE_SELECTED] = Accessibility_STATE_SELECTED;
+  state_table[SPI_STATE_SENSITIVE] = Accessibility_STATE_SENSITIVE;
+  state_table[SPI_STATE_SHOWING] = Accessibility_STATE_SHOWING;
+  state_table[SPI_STATE_SINGLE_LINE] = Accessibility_STATE_SINGLE_LINE;
+  state_table[SPI_STATE_STALE] = Accessibility_STATE_STALE;
+  state_table[SPI_STATE_TRANSIENT] = Accessibility_STATE_TRANSIENT;
+  state_table[SPI_STATE_VERTICAL] = Accessibility_STATE_VERTICAL;
+  state_table[SPI_STATE_VISIBLE] = Accessibility_STATE_VISIBLE;
+
+  return TRUE;
+}
+
+
+
+static Accessibility_StateType
+spi_state_type_from_accessible_state (AccessibleState type)
+{
+  static Accessibility_StateType state_table[Accessibility_STATE_LAST_DEFINED];
+  static SPIBoolean table_initialized = FALSE;
+  Accessibility_StateType rv;
+  
+  if (!table_initialized)
+    table_initialized = init_state_table (state_table);  
+  if (type > SPI_STATE_INVALID && type < SPI_STATE_LAST_DEFINED)
+    rv = state_table[type];
+  else
+    rv = Accessibility_STATE_INVALID;
+  return rv;
+}
+
+
+
+
 /**
  * AccessibleRole_getName:
  * @role: an #AccessibleRole object to query.
@@ -191,7 +256,7 @@ AccessibleRole_getName (AccessibleRole role)
 {
   if (role < MAX_ROLES)
     {
-      return g_strdup (role_names [(int) role]);
+      return CORBA_string_dup (role_names [(int) role]);
     }
   else
     {
@@ -397,9 +462,8 @@ Accessible_getRelationSet (Accessible *obj)
 
   cspi_return_val_if_ev ("getRelationSet", NULL); 
   
-  /* this looks hack-ish, but it's based on the CORBA C bindings spec */
   n_relations = relation_set->_length;
-  relations = malloc (sizeof (AccessibleRelation *) * n_relations);
+  relations = malloc (sizeof (AccessibleRelation *) * (n_relations + 1));
   
   for (i = 0; i < n_relations; ++i)
     {
@@ -475,7 +539,15 @@ Accessible_getRoleName (Accessible *obj)
 AccessibleStateSet *
 Accessible_getStateSet (Accessible *obj)
 {
-  return NULL;
+  AccessibleStateSet *retval;
+
+  cspi_return_val_if_fail (obj != NULL, NULL);
+
+  retval = cspi_object_add (
+    Accessibility_Accessible_getState (CSPI_OBJREF (obj),
+					  cspi_ev ()));
+
+  return retval;
 }
 
 /* Interface query methods */
@@ -1029,11 +1101,13 @@ AccessibleStateSet_contains (AccessibleStateSet *obj,
 			     AccessibleState state)
 {
   CORBA_boolean retval;
-
+  Accessibility_StateType spi_state;
+  
   cspi_return_val_if_fail (obj != NULL, FALSE);
 
+  spi_state = spi_state_type_from_accessible_state (state);
   retval = Accessibility_StateSet_contains (CSPI_OBJREF (obj),
-					    state, cspi_ev ());
+					    spi_state, cspi_ev ());
 
   cspi_return_val_if_ev ("contains", FALSE);
 
@@ -1053,9 +1127,12 @@ void
 AccessibleStateSet_add (AccessibleStateSet *obj,
 			AccessibleState state)
 {
+  Accessibility_StateType spi_state;
+
   cspi_return_if_fail (obj != NULL);
 
-  Accessibility_StateSet_add (CSPI_OBJREF (obj), state, cspi_ev ());
+  spi_state = spi_state_type_from_accessible_state (state);
+  Accessibility_StateSet_add (CSPI_OBJREF (obj), spi_state, cspi_ev ());
   cspi_check_ev ("add");
 }
 
@@ -1072,9 +1149,12 @@ void
 AccessibleStateSet_remove (AccessibleStateSet *obj,
 			   AccessibleState state)
 {
+  Accessibility_StateType spi_state;
+
   cspi_return_if_fail (obj != NULL);
 
-  Accessibility_StateSet_remove (CSPI_OBJREF (obj), state, cspi_ev ());
+  spi_state = spi_state_type_from_accessible_state (state);
+  Accessibility_StateSet_remove (CSPI_OBJREF (obj), spi_state, cspi_ev ());
   cspi_check_ev ("remove");
 }
 
@@ -1127,8 +1207,11 @@ AccessibleStateSet *
 AccessibleStateSet_compare (AccessibleStateSet *obj,
                             AccessibleStateSet *obj2)
 {
+  Accessibility_StateSet retval;
+
   cspi_return_val_if_fail (obj != NULL, NULL);
   cspi_return_val_if_fail (obj2 != NULL, NULL);
+  retval = Accessibility_StateSet_compare (CSPI_OBJREF(obj), CSPI_OBJREF(obj2), cspi_ev ());
   return NULL;	
 }
 
@@ -1145,9 +1228,11 @@ AccessibleStateSet_compare (AccessibleStateSet *obj,
 SPIBoolean
 AccessibleStateSet_isEmpty (AccessibleStateSet *obj)
 {
+  CORBA_boolean retval;
+
   cspi_return_val_if_fail (obj != NULL, FALSE);
-  return TRUE;	
-  /*  return Accessibility_StateSet_isEmpty (CSPI_OBJREF (obj), cspi_ev ());*/
+  retval = Accessibility_StateSet_isEmpty (CSPI_OBJREF (obj), cspi_ev ());
+  return (SPIBoolean) retval;
 }
 
 
